/*
 * button.cc
 *
 * Copyright (C) 2011 Tim Marston <tim@ed.am> and Dan Marston.
 *
 * This file is part of propeller-clock (hereafter referred to as "this
 * program"). See http://ed.am/dev/software/arduino/propeller-clock for more
 * information.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <Arduino.h>
#include <button.h>


Button::Button( int pin )
	:
	_pin( pin ),
	_state( false ),
	_state_millis( 0 ),
	_state_duration_done( 0 ),
	_pending_event( 0 ),
	_send_interim( false ),
	_ignore_next_unpress( false )
{
}


void Button::set_event_times( int event_times[] )
{
	_event_times = event_times;
}


void Button::set_press_mode( bool send_interim )
{
	// if the button is down and we're switching from sending interim presses
	// to not sending interim presses, we need to ignore the next unpress, or
	// this button press may trigger events while being presses *and* events
	// when it is unpressed.
	if( _state && _send_interim && !send_interim )
		_ignore_next_unpress = true;

	// set mode
	_send_interim = send_interim;
}


void Button::update()
{
	// get time
	unsigned long millis = ::millis();

	// get state and check for change
	bool state = digitalRead( _pin )? false : true;
	if( state != _state )
	{
		// if we're not sending interim presses, we care about when the button
		// is being unpressed
		if( !_send_interim && !state )
		{
			// are we being told to ignore the next unpress?
			if( _ignore_next_unpress )
				_ignore_next_unpress = false;
			else
			{
				// check through events to see if the button was pressed long
				// enough to trigger one
				for( int a = 0; _event_times[ a ]; a++ )
				{
					// if this event is in the future, we can stop looking
					if( _state_millis < (unsigned long)_event_times[ a ] )
						break;
					
					// trigger it
					_pending_event = a + 1;
				}
			}
		}
		
		// reset state
		_state = state;
		_state_millis = millis;
		_state_duration_done = 0;
	}

	// if we are sending interim presses, we care about all the time that the
	// button is being pressed
	if( _send_interim && state )
	{
		// calculate new duration for this state
		unsigned long duration = millis - _state_millis;

		// check through events to see if the button has been pressed long
		// enough to trigger one
		for( int a = 0; _event_times[ a ]; a++ )
		{
			// if this event is in the future, we can stop looking
			if( duration < (unsigned long)_event_times[ a ] )
				break;

			// if this event happened since the previous update, trigger it
			if( _state_duration_done < (unsigned long)_event_times[ a ] )
				_pending_event = a + 1;
		}

		// update the duration we've accounted for
		_state_duration_done = duration;
	}
}


int Button::get_event()
{
	int event = _pending_event;
	_pending_event = 0;
	return event;
}
