/*
 * text_renderer.cc
 *
 * Copyright (C) 2011 Tim Marston <tim@ed.am> and Dan Marston.
 *
 * This file is part of propeller-clock (hereafter referred to as "this
 * program"). See http://ed.am/dev/software/arduino/propeller-clock for more
 * information.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "text_renderer.h"
#include "config.h"
#include "Arduino.h"
#include <avr/pgmspace.h>
#include "common.h"
#include "nvram.h"


// cached glyph
static char _glyph_cache[ 8 ];

// the character the current cached glyph is for
static char _glyph_cache_char;

// selected font
static int _font = 0;

// output buffer
static unsigned char _output_buffer;

// does buffer need rendering?
static bool _need_render;


// cache a glyph
void cache_glyph( char c )
{
	static char fonts[][ 66 * 8 ] PROGMEM = {
		{ // INVD-A
			0x00, 0x7c, 0x7e, 0x12, 0x12, 0x12, 0x7e, 0x7c, // A
			0x00, 0x7e, 0x7e, 0x4a, 0x4a, 0x4a, 0x7e, 0x34, // B
			0x00, 0x3c, 0x7e, 0x42, 0x42, 0x42, 0x66, 0x24, // C
			0x00, 0x7e, 0x7e, 0x42, 0x42, 0x66, 0x3c, 0x18, // D
			0x00, 0x7e, 0x7e, 0x4a, 0x4a, 0x4a, 0x42, 0x00, // E
			0x00, 0x7e, 0x7e, 0x0a, 0x0a, 0x0a, 0x02, 0x00, // F
			0x00, 0x3c, 0x7e, 0x42, 0x52, 0x52, 0x76, 0x34, // G
			0x00, 0x7e, 0x7e, 0x08, 0x08, 0x08, 0x7e, 0x7e, // H
			0x00, 0x42, 0x42, 0x7e, 0x7e, 0x42, 0x42, 0x00, // I
			0x00, 0x30, 0x70, 0x40, 0x40, 0x40, 0x7e, 0x3e, // J
			0x00, 0x7e, 0x7e, 0x08, 0x1c, 0x36, 0x62, 0x40, // K
			0x00, 0x7e, 0x7e, 0x40, 0x40, 0x40, 0x40, 0x00, // L
			0x00, 0x7e, 0x7e, 0x0c, 0x18, 0x0c, 0x7e, 0x7e, // M
			0x00, 0x7e, 0x7e, 0x0c, 0x18, 0x30, 0x7e, 0x7e, // N
			0x00, 0x3c, 0x7e, 0x42, 0x42, 0x42, 0x7e, 0x3c, // O
			0x00, 0x7e, 0x7e, 0x12, 0x12, 0x12, 0x1e, 0x0c, // P
			0x00, 0x3c, 0x7e, 0x52, 0x52, 0x62, 0x7e, 0x3c, // Q
			0x00, 0x7e, 0x7e, 0x12, 0x12, 0x32, 0x7e, 0x4c, // R
			0x00, 0x24, 0x6e, 0x4a, 0x4a, 0x4a, 0x7a, 0x30, // S
			0x00, 0x02, 0x02, 0x7e, 0x7e, 0x02, 0x02, 0x00, // T
			0x00, 0x3e, 0x7e, 0x40, 0x40, 0x40, 0x7e, 0x3e, // U
			0x00, 0x1e, 0x3e, 0x60, 0x40, 0x60, 0x3e, 0x1e, // V
			0x00, 0x3e, 0x7e, 0x60, 0x30, 0x60, 0x7e, 0x3e, // W
			0x00, 0x42, 0x66, 0x3c, 0x18, 0x3c, 0x66, 0x42, // X
			0x02, 0x06, 0x0c, 0x78, 0x78, 0x0c, 0x06, 0x02, // Y
			0x00, 0x42, 0x62, 0x72, 0x5a, 0x4e, 0x46, 0x42, // Z
			0x00, 0x20, 0x74, 0x54, 0x54, 0x7c, 0x78, 0x00, // a
			0x00, 0x00, 0x7e, 0x7e, 0x48, 0x48, 0x78, 0x30, // b
			0x00, 0x00, 0x38, 0x7c, 0x44, 0x44, 0x00, 0x00, // c
			0x00, 0x30, 0x78, 0x48, 0x48, 0x7e, 0x7e, 0x00, // d
			0x00, 0x38, 0x7c, 0x54, 0x54, 0x5c, 0x08, 0x00, // e
			0x00, 0x00, 0x7c, 0x7e, 0x0a, 0x02, 0x00, 0x00, // f
			0x00, 0x18, 0xbc, 0xa4, 0xa4, 0xfc, 0x7c, 0x00, // g
			0x00, 0x7e, 0x7e, 0x08, 0x08, 0x78, 0x70, 0x00, // h
			0x00, 0x00, 0x48, 0x7a, 0x7a, 0x40, 0x00, 0x00, // i
			0x00, 0x00, 0x40, 0xc0, 0x80, 0xfa, 0x7a, 0x00, // j
			0x00, 0x00, 0x7e, 0x7e, 0x10, 0x3c, 0x64, 0x40, // k
			0x00, 0x00, 0x00, 0x3e, 0x7e, 0x40, 0x40, 0x00, // l
			0x7c, 0x7c, 0x04, 0x78, 0x7c, 0x04, 0x7c, 0x78, // m
			0x00, 0x7c, 0x7c, 0x04, 0x04, 0x7c, 0x78, 0x00, // n
			0x00, 0x38, 0x7c, 0x44, 0x44, 0x7c, 0x38, 0x00, // o
			0x00, 0xfc, 0xfc, 0x24, 0x24, 0x3c, 0x18, 0x00, // p
			0x00, 0x18, 0x3c, 0x24, 0x24, 0xfc, 0xfc, 0x80, // q
			0x00, 0x00, 0x78, 0x7c, 0x04, 0x04, 0x00, 0x00, // r
			0x00, 0x48, 0x5c, 0x54, 0x54, 0x74, 0x20, 0x00, // s
			0x00, 0x04, 0x3e, 0x7e, 0x44, 0x44, 0x00, 0x00, // t
			0x00, 0x3c, 0x7c, 0x40, 0x40, 0x7c, 0x3c, 0x00, // u
			0x00, 0x0c, 0x3c, 0x70, 0x60, 0x3c, 0x0c, 0x00, // v
			0x3c, 0x7c, 0x40, 0x30, 0x40, 0x7c, 0x3c, 0x00, // w
			0x00, 0x44, 0x6c, 0x38, 0x38, 0x6c, 0x44, 0x00, // x
			0x00, 0x1c, 0xbc, 0xa0, 0xa0, 0xfc, 0x7c, 0x00, // y
			0x00, 0x44, 0x64, 0x74, 0x5c, 0x4c, 0x44, 0x00, // z
			0x00, 0x3c, 0x7e, 0x52, 0x4a, 0x46, 0x7e, 0x3c, // 0
			0x00, 0x00, 0x44, 0x44, 0x7e, 0x7e, 0x40, 0x40, // 1
			0x00, 0x64, 0x76, 0x52, 0x52, 0x52, 0x5e, 0x4c, // 2
			0x00, 0x24, 0x66, 0x42, 0x4a, 0x4a, 0x7e, 0x34, // 3
			0x00, 0x30, 0x28, 0x24, 0x7e, 0x7e, 0x20, 0x20, // 4
			0x00, 0x2e, 0x6e, 0x4a, 0x4a, 0x4a, 0x7a, 0x32, // 5
			0x00, 0x3c, 0x7e, 0x4a, 0x4a, 0x4a, 0x7a, 0x30, // 6
			0x00, 0x02, 0x02, 0x62, 0x72, 0x1a, 0x0e, 0x06, // 7
			0x00, 0x34, 0x7e, 0x4a, 0x4a, 0x4a, 0x7e, 0x34, // 8
			0x00, 0x0c, 0x5e, 0x52, 0x52, 0x52, 0x7e, 0x3c, // 9
			0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, // -
			0x00, 0x00, 0x00, 0x6c, 0x6c, 0x00, 0x00, 0x00, // :
			0x00, 0x00, 0x00, 0x60, 0x60, 0x00, 0x00, 0x00, // .
			0x00, 0x00, 0x00, 0x80, 0xe0, 0x60, 0x00, 0x00, // ,
		}, { // INVD-D
			0x00, 0x7f, 0x7f, 0x09, 0x09, 0x09, 0x7f, 0x7f, // A
			0x00, 0x7f, 0x7f, 0x49, 0x49, 0x49, 0x7f, 0x77, // B
			0x00, 0x7f, 0x7f, 0x41, 0x41, 0x41, 0x77, 0x77, // C
			0x00, 0x7f, 0x7f, 0x41, 0x41, 0x41, 0x7f, 0x3e, // D
			0x00, 0x7f, 0x7f, 0x49, 0x49, 0x49, 0x63, 0x63, // E
			0x00, 0x7f, 0x7f, 0x09, 0x09, 0x09, 0x03, 0x03, // F
			0x00, 0x7f, 0x7f, 0x41, 0x49, 0x49, 0x7b, 0x7b, // G
			0x00, 0x7f, 0x7f, 0x08, 0x08, 0x08, 0x7f, 0x7f, // H
			0x00, 0x41, 0x41, 0x7f, 0x7f, 0x41, 0x41, 0x00, // I
			0x00, 0x60, 0x60, 0x40, 0x40, 0x40, 0x7f, 0x7f, // J
			0x00, 0x7f, 0x7f, 0x08, 0x08, 0x08, 0x7f, 0x77, // K
			0x00, 0x7f, 0x7f, 0x40, 0x40, 0x40, 0x60, 0x60, // L
			0x00, 0x7f, 0x7f, 0x02, 0x04, 0x02, 0x7f, 0x7f, // M
			0x00, 0x7f, 0x7f, 0x06, 0x0c, 0x18, 0x7f, 0x7f, // N
			0x00, 0x7f, 0x7f, 0x41, 0x41, 0x41, 0x7f, 0x7f, // O
			0x00, 0x7f, 0x7f, 0x09, 0x09, 0x09, 0x0f, 0x0f, // P
			0x00, 0x7f, 0x7f, 0x41, 0x51, 0x21, 0x7f, 0x7f, // Q
			0x00, 0x7f, 0x7f, 0x09, 0x09, 0x09, 0x7f, 0x77, // R
			0x00, 0x6f, 0x6f, 0x49, 0x49, 0x49, 0x7b, 0x7b, // S
			0x00, 0x01, 0x01, 0x7f, 0x7f, 0x01, 0x01, 0x00, // T
			0x00, 0x7f, 0x7f, 0x40, 0x40, 0x40, 0x7f, 0x7f, // U
			0x00, 0x1f, 0x3f, 0x60, 0x40, 0x60, 0x3f, 0x1f, // V
			0x00, 0x7f, 0x7f, 0x20, 0x10, 0x20, 0x7f, 0x7f, // W
			0x00, 0x77, 0x7f, 0x08, 0x08, 0x08, 0x7f, 0x77, // X
			0x00, 0x6f, 0x6f, 0x48, 0x48, 0x48, 0x7f, 0x7f, // Y
			0x00, 0x63, 0x73, 0x59, 0x4d, 0x47, 0x63, 0x61, // Z
			0x00, 0x74, 0x74, 0x54, 0x54, 0x7c, 0x7c, 0x00, // a
			0x00, 0x7f, 0x7f, 0x44, 0x44, 0x7c, 0x7c, 0x00, // b
			0x00, 0x7c, 0x7c, 0x44, 0x44, 0x6c, 0x6c, 0x00, // c
			0x00, 0x7c, 0x7c, 0x44, 0x44, 0x7f, 0x7f, 0x00, // d
			0x00, 0x7c, 0x7c, 0x54, 0x54, 0x5c, 0x5c, 0x00, // e
			0x00, 0x00, 0x04, 0x7f, 0x7f, 0x05, 0x05, 0x01, // f
			0x00, 0xbc, 0xbc, 0xa4, 0xa4, 0xfc, 0xfc, 0x00, // g
			0x00, 0x7f, 0x7f, 0x04, 0x04, 0x7c, 0x7c, 0x00, // h
			0x00, 0x44, 0x44, 0x7d, 0x7d, 0x40, 0x40, 0x00, // i
			0x00, 0x80, 0x80, 0xfd, 0xfd, 0x00, 0x00, 0x00, // j
			0x00, 0x7f, 0x7f, 0x10, 0x10, 0x7c, 0x6c, 0x00, // k
			0x00, 0x00, 0x00, 0x7f, 0x7f, 0x40, 0x40, 0x00, // l
			0x00, 0x7c, 0x7c, 0x04, 0x7c, 0x04, 0x7c, 0x7c, // m
			0x00, 0x7c, 0x7c, 0x04, 0x04, 0x7c, 0x7c, 0x00, // n
			0x00, 0x7c, 0x7c, 0x44, 0x44, 0x7c, 0x7c, 0x00, // o
			0x00, 0xfc, 0xfc, 0x24, 0x24, 0x3c, 0x3c, 0x00, // p
			0x00, 0x3c, 0x3c, 0x24, 0x24, 0xfc, 0xfc, 0x00, // q
			0x00, 0x7c, 0x7c, 0x04, 0x04, 0x0c, 0x0c, 0x00, // r
			0x00, 0x5c, 0x5c, 0x54, 0x54, 0x74, 0x74, 0x00, // s
			0x00, 0x00, 0x04, 0x7f, 0x7f, 0x44, 0x44, 0x00, // t
			0x00, 0x7c, 0x7c, 0x40, 0x40, 0x7c, 0x7c, 0x00, // u
			0x00, 0x1c, 0x3c, 0x60, 0x60, 0x3c, 0x1c, 0x00, // v
			0x00, 0x7c, 0x7c, 0x40, 0x7c, 0x40, 0x7c, 0x7c, // w
			0x00, 0x6c, 0x7c, 0x10, 0x10, 0x7c, 0x6c, 0x00, // x
			0x00, 0xbc, 0xbc, 0xa0, 0xa0, 0xfc, 0xfc, 0x00, // y
			0x00, 0x64, 0x74, 0x54, 0x54, 0x5c, 0x4c, 0x00, // z
			0x00, 0x7f, 0x7f, 0x51, 0x49, 0x45, 0x7f, 0x7f, // 0
			0x00, 0x41, 0x41, 0x7f, 0x7f, 0x40, 0x40, 0x00, // 1
			0x00, 0x7b, 0x7b, 0x49, 0x49, 0x49, 0x4f, 0x4f, // 2
			0x00, 0x63, 0x63, 0x49, 0x49, 0x49, 0x7f, 0x7f, // 3
			0x00, 0x1f, 0x1f, 0x10, 0x10, 0x7f, 0x7f, 0x10, // 4
			0x00, 0x6f, 0x6f, 0x49, 0x49, 0x49, 0x79, 0x79, // 5
			0x00, 0x7f, 0x7f, 0x49, 0x49, 0x49, 0x7b, 0x7b, // 6
			0x00, 0x03, 0x03, 0x71, 0x79, 0x0d, 0x07, 0x03, // 7
			0x00, 0x7f, 0x7f, 0x49, 0x49, 0x49, 0x7f, 0x7f, // 8
			0x00, 0x6f, 0x6f, 0x49, 0x49, 0x49, 0x7f, 0x7f, // 9
			0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, // -
			0x00, 0x00, 0x00, 0x66, 0x66, 0x00, 0x00, 0x00, // :
			0x00, 0x00, 0x00, 0x60, 0x60, 0x00, 0x00, 0x00, // .
			0x00, 0x00, 0x80, 0xe0, 0x60, 0x00, 0x00, 0x00, // ,
		}, { // INVD-G
			0x7c, 0x7e, 0x22, 0x7e, 0x7e, 0x7e, 0x7c, 0x00, // A
			0x7c, 0x7e, 0x7e, 0x7e, 0x4a, 0x7e, 0x34, 0x00, // B
			0x3c, 0x7e, 0x7e, 0x7e, 0x42, 0x76, 0x34, 0x00, // C
			0x7c, 0x7e, 0x7e, 0x7e, 0x42, 0x7e, 0x3c, 0x00, // D
			0x3c, 0x7e, 0x7e, 0x7e, 0x4a, 0x4a, 0x00, 0x00, // E
			0x7c, 0x7e, 0x7e, 0x7e, 0x12, 0x12, 0x00, 0x00, // F
			0x3c, 0x7e, 0x7e, 0x7e, 0x52, 0x76, 0x34, 0x00, // G
			0x7e, 0x7e, 0x7e, 0x7e, 0x08, 0x7e, 0x7e, 0x00, // H
			0x00, 0x42, 0x7e, 0x7e, 0x7e, 0x7e, 0x42, 0x00, // I
			0x30, 0x70, 0x42, 0x7e, 0x7e, 0x7e, 0x3e, 0x00, // J
			0x7e, 0x7e, 0x7e, 0x7e, 0x08, 0x76, 0x62, 0x00, // K
			0x00, 0x3e, 0x7e, 0x7e, 0x7e, 0x40, 0x40, 0x00, // L
			0x7e, 0x7e, 0x7e, 0x7c, 0x08, 0x7c, 0x7e, 0x00, // M
			0x7e, 0x7e, 0x7e, 0x7c, 0x08, 0x7e, 0x7e, 0x00, // N
			0x3c, 0x7e, 0x7e, 0x7e, 0x42, 0x7e, 0x3c, 0x00, // O
			0x7c, 0x7e, 0x7e, 0x7e, 0x12, 0x1e, 0x0c, 0x00, // P
			0x3c, 0x7e, 0x7e, 0x7e, 0x42, 0xfe, 0xbc, 0x00, // Q
			0x7c, 0x7e, 0x7e, 0x7e, 0x12, 0x7e, 0x6c, 0x00, // R
			0x24, 0x6e, 0x5e, 0x7e, 0x7a, 0x76, 0x24, 0x00, // S
			0x00, 0x02, 0x7e, 0x7e, 0x7e, 0x7e, 0x02, 0x00, // T
			0x3e, 0x7e, 0x7e, 0x7e, 0x40, 0x7e, 0x3e, 0x00, // U
			0x1e, 0x3e, 0x7e, 0x7e, 0x40, 0x3e, 0x1e, 0x00, // V
			0x3e, 0x7e, 0x7e, 0x7e, 0x20, 0x7e, 0x3e, 0x00, // W
			0x42, 0x66, 0x3e, 0x7e, 0x7c, 0x66, 0x42, 0x00, // X
			0x06, 0x0e, 0x7e, 0x7e, 0x70, 0x0e, 0x06, 0x00, // Y
			0x62, 0x72, 0x7a, 0x7e, 0x5e, 0x4e, 0x46, 0x00, // Z
			0x38, 0x7c, 0x44, 0x3c, 0x7c, 0x7c, 0x78, 0x00, // a
			0x3f, 0x7f, 0x7f, 0x7f, 0x44, 0x7c, 0x38, 0x00, // b
			0x38, 0x7c, 0x7c, 0x7c, 0x44, 0x6c, 0x28, 0x00, // c
			0x38, 0x7c, 0x44, 0x3f, 0x7f, 0x7f, 0x7f, 0x00, // d
			0x38, 0x7c, 0x7c, 0x7c, 0x54, 0x5c, 0x28, 0x00, // e
			0x88, 0xfe, 0xff, 0xff, 0x7f, 0x09, 0x0a, 0x00, // f
			0x58, 0xbc, 0xa4, 0xfc, 0xfc, 0xfc, 0x78, 0x00, // g
			0x7f, 0x7f, 0x7f, 0x7f, 0x04, 0x7c, 0x78, 0x00, // h
			0x00, 0x00, 0x7a, 0x7f, 0x7f, 0x7a, 0x00, 0x00, // i
			0x00, 0x80, 0xfa, 0xff, 0xff, 0x7a, 0x00, 0x00, // j
			0x7f, 0x7f, 0x7f, 0x7f, 0x24, 0x7c, 0x58, 0x00, // k
			0x00, 0x00, 0x3f, 0x7f, 0x7f, 0x7f, 0x40, 0x00, // l
			0x78, 0x7c, 0x7c, 0x04, 0x7c, 0x04, 0x7c, 0x78, // m
			0x7c, 0x7c, 0x7c, 0x78, 0x04, 0x7c, 0x78, 0x00, // n
			0x38, 0x7c, 0x7c, 0x7c, 0x44, 0x7c, 0x38, 0x00, // o
			0xfc, 0xfc, 0xfc, 0xf8, 0x44, 0x7c, 0x38, 0x00, // p
			0x38, 0x7c, 0x44, 0xfc, 0xfc, 0xfc, 0xf8, 0x00, // q
			0x7c, 0x7c, 0x7c, 0x7c, 0x04, 0x0c, 0x08, 0x00, // r
			0x28, 0x5c, 0x5c, 0x7c, 0x74, 0x74, 0x28, 0x00, // s
			0x04, 0x3e, 0x7e, 0x7e, 0x7e, 0x44, 0x24, 0x00, // t
			0x3c, 0x7c, 0x40, 0x3c, 0x7c, 0x7c, 0x7c, 0x00, // u
			0x1c, 0x3c, 0x7c, 0x7c, 0x60, 0x3c, 0x1c, 0x00, // v
			0x3c, 0x7c, 0x7c, 0x40, 0x7c, 0x40, 0x7c, 0x3c, // w
			0x44, 0x6c, 0x3c, 0x7c, 0x78, 0x6c, 0x44, 0x00, // x
			0x9c, 0xbc, 0xfc, 0xfc, 0x60, 0x3c, 0x1c, 0x00, // y
			0x64, 0x74, 0x3c, 0x7c, 0x5c, 0x4c, 0x24, 0x00, // z
			0x3c, 0x7e, 0x42, 0x7e, 0x7e, 0x7e, 0x3c, 0x00, // 0
			0x00, 0x42, 0x7e, 0x7e, 0x7e, 0x7e, 0x40, 0x00, // 1
			0x64, 0x72, 0x7a, 0x7e, 0x5e, 0x4e, 0x44, 0x00, // 2
			0x24, 0x66, 0x4a, 0x7e, 0x7e, 0x7e, 0x34, 0x00, // 3
			0x0e, 0x1e, 0x10, 0x7e, 0x7e, 0x7e, 0x7e, 0x00, // 4
			0x2e, 0x6e, 0x4e, 0x7a, 0x7a, 0x7a, 0x32, 0x00, // 5
			0x3c, 0x7e, 0x7e, 0x7e, 0x4a, 0x7a, 0x34, 0x00, // 6
			0x02, 0x62, 0x7a, 0x7e, 0x7e, 0x1e, 0x06, 0x00, // 7
			0x34, 0x7e, 0x4e, 0x7e, 0x7a, 0x7e, 0x34, 0x00, // 8
			0x2c, 0x5e, 0x52, 0x7e, 0x7e, 0x7e, 0x3c, 0x00, // 9
			0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, // -
			0x00, 0x00, 0x28, 0x7c, 0x7c, 0x28, 0x00, 0x00, // :
			0x00, 0x00, 0x30, 0x78, 0x78, 0x30, 0x00, 0x00, // .
			0x00, 0x00, 0x10, 0xb8, 0xf8, 0x70, 0x00, 0x00, // ,
		}, { // INVD-B
			0x00, 0x78, 0x7f, 0x09, 0x09, 0x0f, 0x78, 0x00, // A
			0x00, 0x7f, 0x79, 0x49, 0x49, 0x4f, 0x78, 0x00, // B
			0x00, 0x7f, 0x79, 0x41, 0x41, 0x41, 0x63, 0x00, // C
			0x00, 0x7f, 0x79, 0x41, 0x41, 0x41, 0x7f, 0x00, // D
			0x00, 0x00, 0x7f, 0x79, 0x49, 0x49, 0x41, 0x00, // E
			0x00, 0x00, 0x7f, 0x79, 0x09, 0x09, 0x01, 0x00, // F
			0x00, 0x7f, 0x79, 0x41, 0x41, 0x49, 0x7b, 0x00, // G
			0x00, 0x7f, 0x78, 0x08, 0x08, 0x08, 0x7f, 0x00, // H
			0x00, 0x00, 0x00, 0x7f, 0x78, 0x00, 0x00, 0x00, // I
			0x00, 0x60, 0x40, 0x40, 0x40, 0x7f, 0x78, 0x00, // J
			0x00, 0x7f, 0x78, 0x08, 0x08, 0x0f, 0x78, 0x00, // K
			0x00, 0x00, 0x7f, 0x78, 0x40, 0x40, 0x40, 0x00, // L
			0x00, 0x7f, 0x79, 0x01, 0x7f, 0x01, 0x7f, 0x00, // M
			0x00, 0x7f, 0x79, 0x01, 0x01, 0x01, 0x7f, 0x00, // N
			0x00, 0x7f, 0x41, 0x41, 0x41, 0x43, 0x7f, 0x00, // O
			0x00, 0x7f, 0x79, 0x09, 0x09, 0x09, 0x0f, 0x00, // P
			0x00, 0x7f, 0x41, 0x41, 0x61, 0x61, 0x7f, 0x00, // Q
			0x00, 0x7f, 0x79, 0x09, 0x09, 0x0f, 0x78, 0x00, // R
			0x00, 0x6f, 0x49, 0x49, 0x49, 0x79, 0x7b, 0x00, // S
			0x00, 0x01, 0x01, 0x7f, 0x79, 0x01, 0x01, 0x00, // T
			0x00, 0x7f, 0x78, 0x40, 0x40, 0x40, 0x7f, 0x00, // U
			0x00, 0x0f, 0x7f, 0x40, 0x40, 0x78, 0x0f, 0x00, // V
			0x00, 0x7f, 0x78, 0x40, 0x7f, 0x40, 0x7f, 0x00, // W
			0x00, 0x77, 0x78, 0x08, 0x08, 0x08, 0x77, 0x00, // X
			0x00, 0x0f, 0x08, 0x78, 0x78, 0x08, 0x0f, 0x00, // Y
			0x00, 0x7b, 0x79, 0x49, 0x49, 0x49, 0x6f, 0x00, // Z
			0x00, 0x74, 0x74, 0x54, 0x54, 0x7c, 0x00, 0x00, // a
			0x00, 0x7f, 0x74, 0x44, 0x44, 0x7c, 0x00, 0x00, // b
			0x00, 0x7c, 0x74, 0x44, 0x44, 0x6c, 0x00, 0x00, // c
			0x00, 0x7c, 0x74, 0x44, 0x44, 0x7f, 0x00, 0x00, // d
			0x00, 0x7c, 0x74, 0x54, 0x54, 0x5c, 0x00, 0x00, // e
			0x00, 0x00, 0x04, 0x7f, 0x75, 0x01, 0x00, 0x00, // f
			0x00, 0xbc, 0xb4, 0xa4, 0xa4, 0xfc, 0x00, 0x00, // g
			0x00, 0x7f, 0x74, 0x04, 0x04, 0x7c, 0x00, 0x00, // h
			0x00, 0x00, 0x00, 0x7d, 0x71, 0x00, 0x00, 0x00, // i
			0x00, 0x00, 0x80, 0xbd, 0xf1, 0x00, 0x00, 0x00, // j
			0x00, 0x7f, 0x70, 0x10, 0x1c, 0x70, 0x00, 0x00, // k
			0x00, 0x00, 0x00, 0x7f, 0x78, 0x00, 0x00, 0x00, // l
			0x00, 0x7c, 0x74, 0x04, 0x7c, 0x04, 0x7c, 0x00, // m
			0x00, 0x7c, 0x74, 0x04, 0x04, 0x7c, 0x00, 0x00, // n
			0x00, 0x7c, 0x74, 0x44, 0x44, 0x7c, 0x00, 0x00, // o
			0x00, 0xfc, 0x34, 0x24, 0x24, 0x3c, 0x00, 0x00, // p
			0x00, 0x3c, 0x34, 0x24, 0x24, 0xfc, 0x00, 0x00, // q
			0x00, 0x00, 0x7c, 0x74, 0x04, 0x04, 0x00, 0x00, // r
			0x00, 0x5c, 0x54, 0x54, 0x74, 0x74, 0x00, 0x00, // s
			0x00, 0x00, 0x04, 0x7f, 0x74, 0x00, 0x00, 0x00, // t
			0x00, 0x7c, 0x70, 0x40, 0x40, 0x7c, 0x00, 0x00, // u
			0x00, 0x1c, 0x7c, 0x40, 0x70, 0x1c, 0x00, 0x00, // v
			0x00, 0x7c, 0x70, 0x40, 0x7c, 0x40, 0x7c, 0x00, // w
			0x00, 0x6c, 0x70, 0x10, 0x10, 0x6c, 0x00, 0x00, // x
			0x00, 0xbc, 0xb0, 0xa0, 0xa0, 0xfc, 0x00, 0x00, // y
			0x00, 0x74, 0x74, 0x54, 0x54, 0x5c, 0x00, 0x00, // z
			0x00, 0x7f, 0x41, 0x41, 0x41, 0x79, 0x7f, 0x00, // 0
			0x00, 0x00, 0x00, 0x78, 0x7f, 0x00, 0x00, 0x00, // 1
			0x00, 0x7b, 0x79, 0x49, 0x49, 0x49, 0x4f, 0x00, // 2
			0x00, 0x63, 0x49, 0x49, 0x49, 0x7f, 0x78, 0x00, // 3
			0x00, 0x1f, 0x11, 0x11, 0x71, 0x7f, 0x10, 0x00, // 4
			0x00, 0x6f, 0x49, 0x49, 0x49, 0x79, 0x79, 0x00, // 5
			0x00, 0x7f, 0x49, 0x49, 0x49, 0x79, 0x7b, 0x00, // 6
			0x00, 0x01, 0x01, 0x01, 0x79, 0x7f, 0x00, 0x00, // 7
			0x00, 0x78, 0x4f, 0x49, 0x49, 0x7f, 0x78, 0x00, // 8
			0x00, 0x1f, 0x11, 0x11, 0x11, 0x71, 0x7f, 0x00, // 9
			0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, // -
			0x00, 0x00, 0x00, 0x6c, 0x6c, 0x00, 0x00, 0x00, // :
			0x00, 0x00, 0x00, 0x60, 0x60, 0x00, 0x00, 0x00, // .
			0x00, 0x00, 0x00, 0x60, 0xe0, 0x00, 0x00, 0x00, // ,
		}
	};

	// nothing to do?
	if( c == _glyph_cache_char )
		return;
	_glyph_cache_char = c;

	// find glyph data position
	int pos;
	if( c >= 'A' && c <= 'Z' )
		pos = c - 'A';
	else if( c >= 'a' && c <= 'z' )
		pos = c - 'a' + 26;
	else if( c >= '0' && c <= '9' )
		pos = c - '0' + 52;
	else if( c == '-' )
		pos = 62;
	else if( c == ':' )
		pos = 63;
	else if( c == '.' )
		pos = 64;
	else if( c == ',' )
		pos = 65;
	else
		pos = -1;

	// space? or copy glyph from progmem?
	if( pos == -1 )
		memset( _glyph_cache, 0, 8 );
	else
		memcpy_P( _glyph_cache, &( fonts[ _font ][ pos * 8 ] ), 8 );
}


void TextRenderer::init()
{
	Nvram::load( Nvram::NVRAM_FONT, _font );
	if( _font < 0 || _font >= 4 ) _font = 0;
}


int TextRenderer::get_width( int message_len, int scale )
{
	return message_len * 8 * scale;
}


void TextRenderer::render( const char *message, int message_len, int x,
						   bool y_flip, int y_shift, int scale )
{
	char glyph_col = 0;

	// handle negative x
	if( x >= 0 )
	{
		// scale font
		x /= scale;

		int pos = x / 8;
		if( pos < message_len )
		{
			// ensure correct glyph is cached and pull out column of data
			cache_glyph( message[ pos ] );
			glyph_col = _glyph_cache[ x % 8 ];
		}
	}

	// draw a column
	if( y_flip )
		for( int a = 7; a >= 0; a-- ) {
			if( glyph_col & 1 )
				_output_buffer |= 1 << ( a + y_shift );
			glyph_col >>= 1;
		}
	else
		for( int a = 0; a < 8; a++ ) {
			if( glyph_col & 1 )
				_output_buffer |= 1 << ( a + y_shift );
			glyph_col >>= 1;
		}
}


void TextRenderer::reset_buffer()
{
	_output_buffer = 0;
	_need_render = false;
}


void TextRenderer::buffer_in_use()
{
	_need_render = true;
}


void TextRenderer::output_buffer()
{
	if( !_need_render ) return;

	for( int a = 8; a >= 0; a-- ) {
		led( a, ( _output_buffer & 1 )? true : false );
		_output_buffer >>= 1;
	}
	_output_buffer = 0;
}


int TextRenderer::get_font()
{
	return _font;
}


void TextRenderer::inc_font()
{
	if( ++_font >= 4 )
		_font = 0;
	Nvram::save( Nvram::NVRAM_FONT, _font );
}
